/* new_tystream.h: Format of the .nty file. $Revision: 1.14 $
 *
 * (c) 2002, Warren Toomey wkt@tuhs.org.
 * 
 * Released under the Gnu General Public License version 2
 */

/* The normal tystream format used by the TiVo is appalling in terms
 * of trying to do editing. The file format described here preserves
 * the data, but in a way that is easy to perform the editing process.
 *
 * The new tystream file format consists of a number of variable-sized
 * records. Each record starts with the following header.
 */

#define NEW_TY_VERSION	0x6e747901	/* Version this header describes */

struct new_ty_header {
  u_int32_t version;		/* Version of this stream format */
  u_int32_t total_size;		/* Size in octets of the whole record */
  u_int32_t index;		/* Index number of this record */
  u_int32_t next_index;		/* Index number of the next record */
  u_int32_t num_entries;	/* Number of entries in the table below */
};

/* All field entries are stored in big-endian (i.e network) format.
 * The version number is 'n', 't', 'y' followed by the number 1.
 * Future versions, if any, will be numbered 2 upwards.
 *
 * The total size is the size in octets of the whole record, including
 * the header, the table, and the data itself.
 *
 * The index number is the number of this record in the file. Note:
 * index numbers must be monotonically increasing, but are not
 * necessarily contiguous. In other words, numbers 1, 2, 3, 7, 11, 51 ...
 * are valid. This allows us to use the original chunk numbers as
 * index numbers, or simply 0,1,2,3,4 if we are creating a new file.
 *
 * Because the next index number is not predictable, this is also stored
 * in the record. To indicate that there is no next record, the next_index
 * field in the last record is set to 0.
 *
 * The number of entries in the table following the header is num_entries.
 * The table consists of the following entries, again in big-endian format.
 */

#define INVALID_ENTRY	0
#define GOP_ENTRY	1
#define SEQUENCE_ENTRY  2
#define I_ENTRY		3
#define P_ENTRY		4
#define B_ENTRY		5
#define AUDIO_ENTRY	6
#define VIDEO_ENTRY	7

struct new_ty_entry {
  u_int32_t type;		/* Type of entry */
  u_int32_t timestamp;		/* Timestamp for the entry */
  u_int32_t size;		/* Size of data area in octets */
};

/* The type field holds one of the constants defined above. The timestamp
 * is the timestamp from the TiVo recording, in 90kHz units. The size is
 * the size of the data area following the table in octets.
 *
 * Each record (header, table, data area) holds a whole Group of Pictures
 * and their associated audio segments. The video entries are in the usual
 * MPEG stored format, i.e GOP I3 B1 B2 P6 B4 B5 P9 B7 B8 P12 B10 B11.
 *
 * Interspersed between the video records are the audio records whose
 * timestamps fit into the time-period of this Group, i.e whose timestamps
 * are >= the lowest video timestamp and are also <= the highest timestamp.
 * Note that the audio records that relate to the B1 and B2 records are
 * stored before the B1 record.
 *
 * After the table, there is a data area containing the video and audio
 * data in the order given in the table. The size of the data area is
 * the sum of the size entries in the table, and also equal to
 * total_size - num_entries * sizeof(struct new_ty_entry).
 *
 * Internally, we store the table in a linked list of the following type.
 */

struct new_ty_node {
  u_int32_t type;		/* Type of entry */
  u_int32_t timestamp;		/* Timestamp for the entry */
  u_int32_t size;		/* Size of data area in octets */
  u_int8_t  *buf;		/* Pointer to the data area in memory */
  struct new_ty_node *next;	/* Pointer to next node in group of pictures */
};
