/*
 * playitsam: Play, edit & save a TiVo video stream from file or MFS.
 * 
 * client.c: Routines to read chunks or streams from a tyserver. $Revision: 1.1
 * $
 * 
 * (c) 2002, Warren Toomey wkt@tuhs.org.
 * 
 * Released under the Gnu General Public License version 2
 */

#include "defines.h"
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include "old_tystream.h"

/* open a socket to a tcp remote host with the specified port */
int open_socket_to_tyserver(char *inhost, int port)
{
  int type = SOCK_STREAM;
  struct sockaddr_in sock_out;
  int res;
  int nodelay = 1;
  int window = 65535;
  char *host;
  struct hostent *hp;

  host = strdup(inhost);
  if (host == NULL) return (-1);

  res = socket(PF_INET, type, 0);
  if (res == -1) { free(host); return (-1); }

  hp = gethostbyname(host);
  if (!hp) { fprintf(stderr, "unknown host: %s\n", host); return (-1); }

  memcpy(&sock_out.sin_addr, hp->h_addr, hp->h_length);
  sock_out.sin_port = htons(port);
  sock_out.sin_family = PF_INET;

  if (connect(res, (struct sockaddr *) & sock_out, sizeof(sock_out))) {
    fprintf(stderr, "failed to connect to %s\n", inhost);
    close(res); free(host); return (-1);
  }

  setsockopt(res, SOL_SOCKET, SO_RCVBUF, &window, sizeof(int));
  setsockopt(res, IPPROTO_TCP, TCP_NODELAY, &nodelay, sizeof(int));

  free(host); return (res);
}

/*
 * Read a specific chunk from the input stream. Return 1 if ok, 0 if failure
 */
int read_a_chunk(int fd, int chunknum, char *bptr)
{
  int numread = 0;
  int numleft = CHUNK_SIZE;
  char ch, buf[100];

  if (bptr==NULL) return(0);

  /*
   * We need to send a CHNK command, get the answer, and then read the 128K
   * of chunk from the TCP connection.
   */
  sprintf(buf, "CHNK %d\r\n", chunknum);
  numread = write(fd, buf, strlen(buf));
  if (numread < 1) {
    fprintf(stderr, "Failure sending CHNK command\n");
    exit(1);
  }
  /* Read the 3-digit result */
  read(fd, buf, 3); buf[3] = '\0';

  if (strcmp(buf, "200")) {	/* Did not get ok message */
    fprintf(stderr, "Couldn't read chunk from tyserver\n");
    close(fd); exit(1);
  }

  /* Otherwise, read up to \n from tyserver before we get the chunk */
  while (1) {
    read(fd, &ch, 1);
    if (ch == '\n') break;
  }

  /* Loop until we get CHUNK_SIZE bytes of data */
  while (numleft > 0) {
    numread = read(fd, bptr, numleft);
    if (numread < 1) {
      fprintf(stderr, "Failure Reading chunk %d\r\n", chunknum);
      exit(1);
    }
    numleft -= numread;
    bptr += numread;
  }
  return (1);
}


/*
 * Open a network connection to an MFS server for an old tystream, and get
 * the number of chunks in the stream back. Returns the file descriptor of
 * the input stream, or if an error it returns -1.
 */
int open_mfs_oldtystream(char *host, char *tyfile, int *numchunks)
{
  int fd;
  int totalchunks;
  char ch, result[4];

  if ((host == NULL) || (tyfile == NULL) || (numchunks == NULL)) return (-1);

  /* Opne a TCP connection to the host */
  fd = open_socket_to_tyserver(host, 1150);
  if (fd == -1) return (-1);

  /* Now send an OPEN command, and get a 200 reply */
  write(fd, "OPEN ", 5);
  write(fd, tyfile, strlen(tyfile));
  write(fd, "\r\n", 2);
  read(fd, result, 3);
  result[3] = '\0';

  if (strcmp(result, "200")) {	/* Did not get ok message */
    fprintf(stderr, "Couldn't play from tyserver\n");
    close(fd); return (-1);
  }

  /* Read the size of the tystream */
  read(fd, &ch, 1);		/* Read the space */
  totalchunks = 0;
  while (1) {
    read(fd, &ch, 1);		/* Read a digit or space */
    if (ch == ' ') break;
    totalchunks = 10 * totalchunks + (ch - '0');
  }
  fprintf(stderr, "Just got %d chunks back from the tyserver!!!\n", totalchunks
    );

  /* Now read up to \n from tyserver before we get the stream */
  while (1) {
    read(fd, &ch, 1);
    if (ch == '\n') break;
  }
  *numchunks = totalchunks;
  return (fd);
}
