/*
 * convert.c: Convert chunks into MPEG records. $Revision: 1.21 $
 * 
 * (c) 2002, Warren Toomey wkt@tuhs.org.
 * 
 * Released under the Gnu General Public License version 2
 */

#include "defines.h"
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "new_tystream.h"
#include "old_tystream.h"
#include "libplayit.h"

extern int Debug;			/* Debug level, for debug output */
extern u_int32_t TotalChunks;		/* Total # of chunks in stream */
int8_t *read_chunk(u_int32_t chunknum);

					/* This is the list we last sent out */
					/* with build_group */
static struct new_ty_node *LastList=NULL;

					/* If we got up to a sequence header
					 * in the previous group, then this
					 * tells us which one it was. It is
					 * important when dealing with chunks
					 * that have multiple sequence hdrs.
					 */
static int Cached_startrec= 0;
static int Cached_chunk_of_startrec= -1;
static u_int8_t *VidBuf= NULL;		/* Our contiguous video buffer */
static int total_vidcopied= 0;		/* Used for offset into out new buf */


/*
 * Return the timestamp in a 16-byte A or V record, in 90kHz units. 0 is
 * returned if error.
 */
static u_int32_t get_timestamp(u_int8_t * cptr)
{
  unsigned long tim;

  if (cptr[8] != 0x07) return (0);
  if (cptr[9] != 0x21) return (0);
  tim= ((((((((cptr[9] & 0x06) << 7) | cptr[10]) << 7) |
	    (cptr[11] >> 1)) << 8) | cptr[12]) << 7) | (cptr[13] >> 1);
  return (tim);
}

/*
 * Given a Buf pointer and a size of the data, return the offset to the first
 * 00 00 01 triplet in the buffer. Return -1 if not found.
 */
static int find000001(u_int8_t * Buf, int size)
{
  int i;

  if (Buf == NULL) return (-1);
  if (size < 3) return (-1);

  for (i= 0; i < size - 3; i++, Buf++) {
    if ((Buf[0] == 0) && (Buf[1] == 0) && (Buf[2] == 1)) return (i);
  }
  return (-1);
}


/* Debug routine: print out some hex. */
static void print_somehex(u_int8_t * cptr, int num)
{
  int i;
  int total;

  total= 16;
  if (num < total) total= num;

  for (i= 0; i < total; i++, cptr++) {
    printf("%02x ", (*cptr & 0xff));
  }
}


static char *Typechar= "XGSIPBAV";

/*
 * Debug routine: Print out a new_ty_node linked list.
 */
void print_newtylist(struct new_ty_node * This, int num)
{
  u_int32_t milli;
  if (This == NULL) return;

  printf("Chunk %d\n", num);

  for (; This != NULL; This= This->next) {
    milli= This->timestamp / 90;
    printf("%c size %5d time %6d: ", Typechar[This->type], This->size, milli);
    print_somehex(This->buf, This->size);
    printf("\n");
  }
  printf("\n");
}


/* The is a multi-function routine. Here is what we try to do:
 * 1. Find a chunk with a video sequence header which is the
 *    beginning of a group.
 * 2. Find a second chunk with a video sequence header, this will
 *    be the sequence header of the next group. We cache the record
 *    where we found this second sequence header.
 * 3. Find the lowest video timestamp between the sequence headers,
 *    i.e within the group itself.
 * 4. Find the lowest video timestamp after the second sequence
 *    header, as we need this to determine the limits for the
 *    audio timestamps.
 * 5. Find audio records, timestamp them, and build a link list.
 *    We capture all of them, and trim the outlyers away later.
 * 6. Also work how much buffer space we need to malloc() for the
 *    video data, and eventually malloc this buffer.
 *
 * We return the audio linked list, and also the two video timestamps.
 * If the chunknum chunk doesn't have a sequence header, we return NULL.
 */

static struct new_ty_node *parse_audchunks(u_int32_t chunknum, u_int32_t *lowtim, u_int32_t *hitim)
{
  int c, vsh;
  int i, startrec, Numrecs;
  int size,total_size=0;
  int found_top_audio=0;
  int found_next_vsh=0;
  struct chunk_rec *p;
  u_int8_t *buf;
  u_int32_t lowest_vidstamp=0xffffffff, highest_vidstamp=0xffffffff;
  u_int32_t vidstamp= 0, audstamp= 0;
  u_int8_t *recptr;		/* Ptr to data in old chunk */
  u_int8_t *newaudptr;          /* Malloc'd copy of old data */
  struct new_ty_node *Head=NULL, *Last=NULL, *This=NULL;


  for (c=chunknum; c<TotalChunks; c++) {

    if (c==TotalChunks-1) return(NULL);

    /* Read in the chunk */
    buf= read_chunk(c);
    if (buf == NULL) return (NULL);

    /* It appears that if buf[1] is >0, it's a chunk which we
     * should ignore.
     */
    if (buf[1]>0)
	continue;

    /* Calculate the video sequence header record, if any */
    vsh= (int)buf[2] + ((int)buf[3] << 8); vsh &= 0x7fff;

    /* If this is the first chunk, give up if there is no vsh.
     * Use the vsh as the startrec unless we have a cached value.
     */
    if (c==chunknum) {
      if (vsh==0x7fff) return(NULL);
      if (c==Cached_chunk_of_startrec) startrec=Cached_startrec;
      else startrec=vsh;
    }
    /* otherwise always start at record 0 */
    else startrec=0;

    /* Loop over the records in this chunk */
    Numrecs= (int)buf[0];

    p= (struct chunk_rec *) & buf[4];
    recptr= buf + 4 + Numrecs * sizeof(struct chunk_rec);
    for (i= 0; i < Numrecs; i++, p++, recptr += size, Last= This) {
      size= FROM_CHUNK_SIZE(p->size);

      /* If we found a vsh not in the first chunk, it's the next group */
      if ((c!=chunknum) && (i==vsh)) {
	found_next_vsh=1;
	Cached_startrec=vsh;
	Cached_chunk_of_startrec=c;
      }

      switch (p->type) {
      case WEIRD0x1_RECORD:
      case WEIRD0x2_RECORD:
      case WEIRD0x3_RECORD:
      case WEIRD0x4_RECORD:	/* No actual data, don't shift start up */
      case WEIRD0x5_RECORD:
        size= 0; break;

      case AUDIO_RECORD:
	if (size==0) break;
        /*
         * We cache any timestamp if we see one. Not all 16-byte records are
	 * timestamps. They must start with 00 00 01 c0.
         */
        if (size == 16) {
	  if ((recptr[0] == 0x0) && (recptr[1] == 0x0) &&
	      (recptr[2] == 0x1) && (recptr[3] == 0xc0)) {
	    audstamp= get_timestamp(recptr);
	    break;
	  }
        }

	/* If this audio record is above the top video timestamp,
	 * set a flag and breakout of the loop.
	 */
	if (audstamp>highest_vidstamp) { found_top_audio=1; break; }

        /* Copy the audio data to our buffer */
        newaudptr= (u_int8_t *) malloc(size);
        if (newaudptr == NULL) return (NULL);
        memcpy(newaudptr, recptr, size);

        /* Build a node for this audio data */
        This= (struct new_ty_node *) malloc(sizeof(struct new_ty_node));
        if (This == NULL) return (NULL);

        This->size= size;
        This->type= AUDIO_ENTRY;
        This->timestamp= audstamp;
        This->buf= newaudptr;
        This->next= NULL;

        if (Head == NULL) Head= This;
        if (Last != NULL) Last->next= This;
 	break;

	case VIDEO_RECORD:
	/* For video records, we are only interested in the size
	 * and the timestamps for now. Size ony between the vshs,
	 * lowest timestamps immediately following each vsh.
	 */

	if (i<startrec) continue;
	if (!found_next_vsh) total_size+=size;

        /*
         * Not all 16-byte records are timestamps.
	 * They must start with 00 00 01 e0.
         */
	if (size!=16) continue;
	if ((recptr[0] == 0x0) && (recptr[1] == 0x0) &&
	    (recptr[2] == 0x1) && (recptr[3] == 0xe0)) {
	  vidstamp= get_timestamp(recptr);

	  if (found_next_vsh) {		/* update the highest timestamp */
	    if (vidstamp<highest_vidstamp) highest_vidstamp=vidstamp;
	  } else {
	    if (vidstamp<lowest_vidstamp) lowest_vidstamp=vidstamp;
	  }
	}
      }
    }

    /* We can leave the chunk reading loop once we have found
     * a single audio record that's above the highest video timestamp
     */
    if (found_top_audio==1) break;
  }

  /* We now need to malloc the video buffer */
  if (VidBuf != NULL) { free(VidBuf); VidBuf= NULL; }
  VidBuf= (u_int8_t *) malloc(total_size);

  if (Debug>=2) {
    printf("We have just read from chunk %d to chunk %d (%d cache needed)\n",
					chunknum, c, c-chunknum+1);
    printf("Just allocated %d octets for VidBuf\n", total_size);
    printf("Video timestamps are %d and %d\n",
				lowest_vidstamp/90, highest_vidstamp/90);
    printf("Here is what the audio linked list looks like:\n");
    print_newtylist(Head, chunknum);
  }

  *lowtim= lowest_vidstamp; *hitim=highest_vidstamp;
  return(Head);
}



/* Read in the given chunk, and build a struct new_ty_node list which
 * contains just the raw video records, but with timestamps built
 * into the nodes. If stopatseq is 0, we must find a sequence header and
 * start at it. Otherwise, we start at record 0 and go up to the sequence
 * header, if there is one. If we find a sequence header, we set stopnow
 * to 1. We return the linked list, or NULL if we can't build one.
 */
static struct new_ty_node * parse_vidchunk(u_int32_t chunknum, int stopatseq, int *stopnow)
{
  u_int8_t *buf;
  struct new_ty_node *Head= NULL, *Last= NULL, *This= NULL;
  int startrec, endrec, Numrecs;
  int i, vsh, size;
  u_int32_t timestamp= 0;
  struct chunk_rec *p;
  u_int8_t *oldvidptr;		/* Ptr to data in old chunk */
  u_int8_t *newvidptr;		/* Ptr to data in new buffer */

  /*
   * Read the chunk in from the cache or elsewhere, and get a pointer to the
   * input buffer.
   */
  buf= read_chunk(chunknum);
  if (buf == NULL) return (NULL);

  /* It appears that if buf[1] is >0, it's a chunk which we
   * should ignore.
   */
  if (buf[1]>0) { *stopnow=3; return (NULL); }

  /*
   * Determine which chunk records we need to look at. This depends on if we
   * are searching for a GOP to start at, to end at, and also if we are
   * starting from the same chunk that we ended at last time.
   */
  Numrecs= (int)buf[0];

  /* Numrecs can't be zero, give up if it is. */
  if (Numrecs==0) {
    *stopnow=3;
    return(NULL);
  }

  vsh= (int)buf[2] + ((int)buf[3] << 8);
  vsh &= 0x7fff;

  /*
   * Vsh indicates the record where there is a sequence header. If it is
   * 0x7fff, then there is no sequence header. If stopatseq is 0 then we
   * can't find one here, so return NULL.
   */
  if (vsh == 0x7fff) {
    if (stopatseq == 0) return (NULL);

    /* We need to parse the whole chunk then */
    startrec= 0; endrec= Numrecs; *stopnow= 0;
  } else {

    /*
     * There is a sequence header. Do we need to stop at it? Yes if stopatseq
     * is 1, i.e we have something from the previous chunk.
     */
    if (stopatseq == 1) {
      startrec= 0;
      endrec= vsh;
      *stopnow= 1;
      Cached_startrec= vsh;
      Cached_chunk_of_startrec= chunknum;
    } else {

      /*
       * Ok, now we are looking for a sequence header as stopatseq==1 and
       * there is one, as vsh is not 0x7fff. However, we may need to start at
       * the cached startrec, as this is where we got up to in the last
       * group.
       */
      if (chunknum == Cached_chunk_of_startrec) {
	startrec= Cached_startrec;
	endrec= Numrecs; *stopnow= 0;
      } else {
	startrec= vsh;
	endrec= Numrecs; *stopnow= 0;
      }
    }
  }

  if (Debug >= 1)
    printf("Chunk %d, %d recs, vsh is %d, start/end is %d %d\n",
	   chunknum, Numrecs, vsh, startrec, endrec);

  newvidptr= VidBuf + total_vidcopied;

  /*
   * Build a linked list of new_ty_nodes which are for now just VIDEO_ENTRY
   * and with the sizes as given from the chunk.
   */
  p= (struct chunk_rec *) & buf[4];

  oldvidptr= buf + 4 + Numrecs * sizeof(struct chunk_rec);
  for (i= 0; i < endrec; i++, p++, oldvidptr += size, Last= This) {
    size= FROM_CHUNK_SIZE(p->size);

    switch (p->type) {
    case WEIRD0x1_RECORD:
    case WEIRD0x2_RECORD:
    case WEIRD0x3_RECORD:
    case WEIRD0x4_RECORD:	/* No actual data, don't shift start up */
    case WEIRD0x5_RECORD:
      size= 0; break;

    case VIDEO_RECORD:
      if (i < startrec) continue;

      /*
       * Not all 16-byte records are timestamps. They must start with 00 00
       * 01 e0.
       */
      if (size == 16) {
	if ((oldvidptr[0] == 0x0) && (oldvidptr[1] == 0x0) &&
	    (oldvidptr[2] == 0x1) && (oldvidptr[3] == 0xe0)) {
	  timestamp= get_timestamp(oldvidptr);
	  break;
	}
      }
      /* Copy the video data to our buffer */
      memcpy(newvidptr, oldvidptr, size);
      total_vidcopied += size;

      /* Build a node for this video data */
      This= (struct new_ty_node *) malloc(sizeof(struct new_ty_node));
      if (This == NULL) return (NULL);

      This->size= size;
      This->type= VIDEO_ENTRY;
      This->timestamp= timestamp;
      This->buf= newvidptr;
      newvidptr += size;
      This->next= NULL;

      if (Head == NULL) Head= This;
      if (Last != NULL) Last->next= This;
    }
  }
  /* If we return a NULL head here, then it's because vsh was record 0
   * and we wanted to stop before vsh. So we set *stopnow to 2 here.
   */
  if (Head==NULL) *stopnow=2;
  return (Head);
}


/* Free a linked list. Free the individual nodes, and also free the
 * audio buffers.
 */
void free_list(struct new_ty_node * This)
{
  struct new_ty_node *Tail;

  while (This!=NULL) {
    Tail= This->next;
    if (This->type == AUDIO_ENTRY) free(This->buf);
    free(This); This= Tail;
  }
}


/* Given a chunk number, try to build a group of pictures starting at
 * this chunk. We first find a chunk with a sequence header, and we
 * collect a linked list of video records from this chunk up to the
 * next chunk with a sequence header. We determine the lowest and
 * highest video timestamps, and we also build an audio linked list
 * including data from all the chunks read in, plus the ones up
 * until the audio timestamp is bigger than the top video timestamp.
 * We then trim the audio linked list to match the video timestamps.
 * A linked list is returned which is the video list followed by the
 * audio list. If we can't build a group, we return NULL. The nextchunk
 * parameter is set to the chunk where we found the sequence header
 * for the next group after this one.
 */
struct new_ty_node * build_group(u_int32_t chunknum, u_int32_t *nextchunk)
{
  extern u_int32_t TotalChunks;	/* Total number of chunks in the file */
  u_int8_t *vidptr;
  struct new_ty_node *VidHead= NULL, *VidThis, *VidTail=NULL, *VidLast;
  struct new_ty_node *AudHead= NULL, *AudThis, *AudTail, *AudLast;
  u_int32_t lovid_timestamp;
  u_int32_t hivid_timestamp;
  int chk, stop, offset, pictype;
  int stopatseq= 0;

  if (LastList!=NULL) free_list(LastList);

  /* Try to find a chunk with a sequence header starting at chunknum */
  for (chk= chunknum; chk < TotalChunks-1; chk++) {
    AudHead=parse_audchunks(chk, &lovid_timestamp, &hivid_timestamp);
    if (AudHead!=NULL) break;
  }

  /* If we ran out of chunks, give up */
  if (chk == TotalChunks-1) { return (NULL); }

  /*
   * Now loop across the chunks following, getting more of the list until we
   * are told to stop. We also need to walk the existing list until we find
   * the tail, so we can link them.
   */
  for (total_vidcopied= 0; chk < TotalChunks; chk++) {

    VidThis= parse_vidchunk(chk, stopatseq, &stop);

    /* VidThis can come back NULL when the vsh was record 0.
     * If this happens, then stop will be set to two!
     * And if we had to skip a chunk because it had no data, then
     * stop will be set to three!
     * Otherwise a real failure so return NULL.
     */
    if (VidThis == NULL) {
      if (stop == 3) continue;
      if (stop == 2) break;
      return (NULL);
    }
    stopatseq= 1;
    if (VidTail) VidTail->next= VidThis;
    else VidHead=VidTail=VidThis;
    if (stop == 1) break;

    while (VidTail->next != NULL) VidTail= VidTail->next;
  }

  /* If we ran out of chunks, give up */
  if (chk == TotalChunks) { return (NULL); }
  *nextchunk= chk;

  if (Debug>=2) {
    printf("Just before mangling the video data, we have video\n");
    print_newtylist(VidHead, chk);
  }

  /*
   * We now have:
   * - a single VidBuf with a copy of the video data
   * - two linked lists pointing at the records therein
   * 
   * What we don't have is an idea of the video record types, nor their exact
   * position.
   * 
   * We now walk the video list, determine the type of record and shift its
   * pointer and size (and the previous size) in order to correctly set the
   * MPEG node types.
   */
  for (VidLast= VidThis= VidHead; VidThis != NULL; VidLast= VidThis, VidThis= VidThis->next) {

    /* Find the start of an MPEG record */
    offset= find000001(VidThis->buf, 16);

    /*
     * We merge this node with the previous one if we didn't find an offset,
     * or if it extends past this node anyway.
     */
    if ((offset == -1) || (offset >= VidThis->size)) {
      VidLast->next= VidThis->next;
      VidLast->size += VidThis->size;
      free(VidThis);
      VidThis= VidLast;
      continue;
    }
    vidptr= VidThis->buf + offset;

    /* Determine the MPEG type of this node */
    switch (vidptr[3]) {
    case GROUP_START:
      VidThis->type= GOP_ENTRY; VidThis->timestamp= 0; break;
    case SEQUENCE_START:
      VidThis->type= SEQUENCE_ENTRY; VidThis->timestamp= 0; break;
    case PICTURE_START:
      pictype= (vidptr[5] & PICTURE_CODING_MASK) >> PICTURE_CODING_SHIFT;
      switch (pictype) {
      case PICTURE_CODING_INTRA:
	VidThis->type= I_ENTRY; break;
      case PICTURE_CODING_BIDIRECTIONAL:
	VidThis->type= B_ENTRY; break;
      case PICTURE_CODING_PREDICTIVE:
	VidThis->type= P_ENTRY; break;
      default:
	printf("Unknown pictype %d\n", pictype);
	VidThis->type= VIDEO_ENTRY;
      }
      break;
    default:
      /* Unknown/unwanted MPEG type, merge with previous entry */
      VidLast->next= VidThis->next;
      VidLast->size += VidThis->size;
      free(VidThis);
      VidThis= VidLast;
      continue;
    }

    /* Adjust the pointers and sizes of things */
    VidThis->buf += offset;
    VidThis->size -= offset;
    if (VidLast != VidThis) VidLast->size += offset;

    /* Fiddle with P-frame and SYS/GOP timestamps */
    if (VidThis->type == GROUP_START) VidThis->timestamp= 0;
    if (VidThis->type == SEQUENCE_START) VidThis->timestamp= 0;
  }

  /*
   * Now trim off any audio records that fall outside the video
   * timestamps.
   */
  for (AudLast= AudThis= AudHead; AudThis != NULL;) {
    if (AudThis->timestamp < lovid_timestamp) {
      AudHead= AudLast= AudThis->next;
      free(AudThis->buf);
      free(AudThis);
      AudThis= AudLast;
      continue;
    }
    if (AudThis->timestamp >= hivid_timestamp) {
      AudLast->next= NULL;

      /* Now free all these node */
      while (AudThis) {
	AudTail= AudThis->next;
	free(AudThis->buf);
	free(AudThis);
	AudThis= AudTail;
      }
      break;
    }
    AudLast= AudThis;
    AudThis= AudThis->next;
  }
#if 0
  printf("The final result is, we have video\n");
  print_newtylist(VidHead, chk);
  printf("The final result is, we have audio\n");
  print_newtylist(AudHead, chk);
#endif

  /* Assert check: VidBuf cannot be NULL */
  if (VidBuf==NULL) return(NULL);

  /*
   * Finally, join the audio list to the video list before returning it.
   */
  for (VidThis= VidHead; VidThis->next != NULL; VidThis= VidThis->next);
  VidThis->next= AudHead;

  if (Debug>=2) {
    printf("The FINAL result is, we have both\n");
    print_newtylist(VidHead, chk);
  }

  LastList= VidHead;
  return (VidHead);
}
