/*
 * playitsam: Play, edit & save a TiVo video stream from file or MFS.
 *
 * cuts.c: Cut list processing for Playitsam. $Revision: 1.22 $
 *
 * (c) 2002, Warren Toomey wkt@tuhs.org.
 *
 * Released under the Gnu General Public License version 2
 */

#include "defines.h"
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include "cuts.h"
#include "nplayitsam.h"

extern u_int32_t TotalChunks;   /* Total number of chunks in the file */

/* We keep cuts on a singly-linked list ordered by the
 * chunk # of the start of the cut. Yes, we don't deal
 * with adjacent or overlapping cuts. Perhaps you would
 * like to add the code for that?!
 */

struct cutlist {
  int start;
  int end;
  int is_splitpoint;
  struct cutlist *next;
};

static struct cutlist *Cuthead=NULL;	/* We keep the head of the list here */

/* When the user starts a cut, we cannot add a new node to the
 * list immediately. Instead, we record the start here.
 */
static int pending_cutstart=0;

void add_cut_start(int start)
{
  pending_cutstart=start;
}

/* Add a new cut to the list, and return the start chunk #,
 * or if there is some error, return -1.
 * If how==CMD_CEND, mark as an end of a cut.
 * If how==CMD_SPLITPOINT, mark as a splitpoint.
 * If how==CMD_CENDSPLIT, mark as an end and a splitpoint.
 */
int add_cut_end(int end, int how)
{
  struct cutlist *this, *S, *L;
  int start;

  if ((how==CMD_CEND)||(how==CMD_CENDSPLIT)) {
    if (pending_cutstart==NO_PENDING_CUTSTART) return(-1);
    if (end < pending_cutstart) return(-1);
    start=pending_cutstart;
    pending_cutstart= NO_PENDING_CUTSTART;
  } else start=end; 

  this= (struct cutlist *)malloc(sizeof(struct cutlist));
  if (this==NULL) return(-1);

  this->start= start; this->end= end; this->next=NULL;
  if ((how==CMD_SPLITPOINT)||(how==CMD_CENDSPLIT)) this->is_splitpoint=1;
  else this->is_splitpoint=0;

  if (Cuthead==NULL) { Cuthead=this; return(start); }

  if (Cuthead->start > this->start) { 
    this->next= Cuthead; Cuthead= this; return(start);
  }

  for (L=S=Cuthead; S; L=S, S=S->next) {
    if (S->next && (S->next->start >= start) &&
				(S->start <= start)) {
	this->next= S->next; S->next= this; return(start);
    }
  }

  /* If we get here, then we fell off the end. Fortunately, L holds
   * the last valid pointer.
   */
   L->next= this; return(start);
}

/* Find the cut starting at this chunk, delete it
 * and return 0, or if there is some error, return -1.
 */
int kill_cut_start(int start)
{
  struct cutlist *S, *L;

  for (L=S=Cuthead; S; L=S, S=S->next) {
    if (S->start == start) {
      if (Cuthead==S) { Cuthead= S->next; }
      else { L->next= S->next; }
      free(S); return(0);
    }
  }
  return(-1);
}

/* Find the cut ending at this chunk, delete it
 * and return 0, or if there is some error, return -1.
 */
int kill_cut_end(int end)
{
  struct cutlist *S, *L;

  for (L=S=Cuthead; S; L=S, S=S->next) {
    if (S->end == end) {
      if (Cuthead==S) { Cuthead= S->next; }
      else { L->next= S->next; }
      free(S); return(0);
    }
  }
  return(-1);
}

void kill_all_cuts(void)
{
  struct cutlist *S, *L;

  for (S=Cuthead; S; ) { L=S; S=S->next; free(L); }
  Cuthead= NULL;

  pending_cutstart=0;
}

/* Given a chunk number, find out if this chunk is inside
 * a cut, outside, or on the edge. Return one of the IS_CUT_
 * constants. However, if allow_playback is 2, return quickly.
 */
int find_chunk_in_cutlist(int chunk, int *cut_end, int *is_splitpoint)
{
  struct cutlist *S;

  *is_splitpoint=0;

  if (pending_cutstart!=NO_PENDING_CUTSTART)
    if (chunk>=pending_cutstart) { *cut_end= 0x7fffffff; return(IS_CUT_START); }

  for (S=Cuthead; S; S=S->next) {
    if (chunk == S->end) {	
      /* Check if it's a splitpoint, and set the flag */
      if (S->is_splitpoint==1) {
	*is_splitpoint=1;
      }
      return(IS_CUT_END);
    }
    if (chunk == S->start) {
      *cut_end= S->end; return(IS_CUT_START);
    }
    if ((S->start < chunk) && (S->end > chunk)) {
      *cut_end= S->end; return(IS_INSIDE_CUT);
    }
  }
  return(IS_OUTSIDE_CUT);
}

/* Print out the list of cuts.
 */
void list_cuts(int chunknum)
{
  int diff, total=0;
  int highest_end=0;
  
  struct cutlist *S;

  /* Print out the list of cuts. Assume we want the whole
   * stream, and subtract each cut.
   */
  total= TotalChunks;
  for (S=Cuthead; S; S=S->next) {

    if (S->is_splitpoint==1)
	printf("  splitpoint at chunk %d\r\n", S->end);
    if (S->end > S->start) {
        printf("  cut from chunk %d to %d\r\n", S->start, S->end);
        diff= S->end - S->start;
        total-= diff;
	if (S->end > highest_end) highest_end= S->end;
    }
  }

  if (pending_cutstart!=NO_PENDING_CUTSTART) {
    printf("  cut from chunk %d to infinity\r\n",pending_cutstart);

    /* Subtract this cut from the total */
    diff= TotalChunks - pending_cutstart;
    total-= diff;
  }
  printf("Total recording is %d Megs\r\n",total/8);

  if ((chunknum>highest_end) && (pending_cutstart==NO_PENDING_CUTSTART)) {
    diff= TotalChunks - chunknum;
    printf("If you cut here, you would have %d Megs\r\n",(total-diff)/8);
  }
}
