/*
 * Eventually, code will appear here that will be able to multiplex
 * new_ty_node GOPs so that we can write out a single MPEG file.
 * However, this isn't likely to occur in the near future, so for
 * now this file is not compiled into the libplayit.a library.
 */

#include "defines.h"
#include <sys/types.h>
#include <stdio.h>
#include "new_tystream.h"
#include "cuts.h"
#include "libplayit.h"

extern int Debug;               /* Turn on debugging messages */
extern int SeparateFiles;       /* Place material into several files */
extern u_int32_t TotalChunks;   /* Total number of chunks in the file */

/* Because we might have made cuts to the tystream, we can't use the
 * timestamps taken from the original stream. We keep an accumulating
 * timestamp called base_stamp. Timestamps for a record in the group
 * are first calculated relative to the lowest timestamp, and then
 * we add base_stamp on this. Then at the end of each group, we add
 * on one video frame's delay to get the base_stamp up to the next group.
 */
static u_int32_t base_stamp=0;
static FILE *mfile=NULL;
static u_int32_t sector_size=2324;
static u_int32_t SCR=0, SCR_increment=0;
static u_int32_t MuxRate=0;

/* Given a group linked-list, determine the video bitrate
 * from the sequence header, and the audio bitrate. Add
 * them together, add 5%, and return as a rate in
 * 50-byte per second units. Returns 0 on error.
 */
u_int32_t determine_muxrate(struct new_ty_node *H)
{
 struct new_ty_node *A;
 u_int32_t video_bitrate, audio_bitrate, bitrate;

 if (H==NULL) return(0);

 /* Find the first audio record */
 for (A=H; (A!=NULL) && (A->type!=AUDIO_ENTRY); A=A->next) ;
 if (A==NULL) return(0);

 /* First video record must be a sequence entry */
 if (H->type!=SEQUENCE_ENTRY) return(0);

 video_bitrate= ((u_int32_t)H->buf[8]<<10) | ((u_int32_t)H->buf[9]<<2) |
					       ((H->buf[10]>>6) & 0x03);
 audio_bitrate= 480;
 bitrate= video_bitrate + audio_bitrate;
 bitrate= bitrate * 21/20;
 printf("SCR goes up by %d\n",sector_size*1800/bitrate);
 SCR_increment= 1+ (sector_size*1800/bitrate);
 return(bitrate);
}

void write_pack_header(void)
{
  u_int8_t start[14]= { 0, 0, 0, 0xba, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0xf8 };
  int i;

  /* Set the timestamp. We assume that SCR_ext is 0
   * and so is bit 33, the most significant bit.
   */

  start[4]= 0x44 | ((SCR>>27) & 0x38) | ((SCR>>28) & 0x3);
  start[5]= (SCR>>20) & 0xff;
  start[6]= 0x4 | ((SCR>>12) & 0xf8) | ((SCR>>13) & 0x3);
  start[7]= (SCR>>5) & 0xff;
  start[8]= 0x4 | ((SCR<<3) & 0xff);
  start[10]= (MuxRate>>14) & 0xff;
  start[11]= (MuxRate>>6) & 0xff;
  start[12]= 0x3 | ((MuxRate<<2) & 0xff);

  SCR+= SCR_increment;

  printf("Got 0x%08x timestamp 0x%08x bitrate\n",SCR,MuxRate);
#if 0
  for (i=0;i<14;i++) printf("0x%02x ",start[i]);
  printf("\n\n");
#endif
}

void write_ptsdts_timestamp(int pestype, u_int32_t timestamp)
{
}

int write_pes_header(int pestype, int length, int aligned, u_int32_t pts, u_int32_t dts)
{
}

void write_single_record(struct new_ty_node *R, u_int32_t timestamp)
{
  /* Fiddle around with the timestamps for P- and I-frames */
}

int write_group_as_mpeg(struct new_ty_node *V)
{
  struct new_ty_node *A=NULL;
  u_int32_t lowtim=0xffffffff;
  u_int32_t hitim=0;
  u_int32_t t;

  if (V==NULL) return(0);

  /* We're only doing this once, but for VBR we should
   * do it for every group.
   */
  if (MuxRate==0) MuxRate=determine_muxrate(V);

  /* Find the lowest and highest video timestamps in this group.
   * Also find the audio head, and determine the video bitrate
   * for this group.
   */
  for (A=V;A!=NULL;A=A->next) {
    if (A->type==AUDIO_ENTRY) break;

    if (A->timestamp<lowtim) lowtim= A->timestamp;
    if (A->timestamp>hitim)  hitim=  A->timestamp;
  }


  /* Loop while we have any records to process */
}

int open_mpeg_file(char *fname)
{
  mfile= fopen(fname, "wb");
  if (mfile==NULL) return(0);
  base_stamp=0;
  return(1);
}

int write_mpeg_file(char *mpeg_name)
{
  int i, c, chunk, cutend;
  int is_splitpoint;
  struct new_ty_node *H;
  int file_has_data = 0;
  int groups = 0;

  if (open_mpeg_file(mpeg_name)==0) return(0);
  if (SeparateFiles != 0) SeparateFiles++;

  /* Loop over all the chunks */
  for (chunk = 0; chunk < TotalChunks;) {

    /* Determine what to do with this chunk number */
    i = find_chunk_in_cutlist(chunk, &cutend, &is_splitpoint);

    switch (i) {
    case IS_INSIDE_CUT:
    case IS_CUT_START:
      chunk = cutend; continue;
    }

    /* We will try to read c in */
    c = chunk;

    if (Debug >= 1) printf("Getting GOP starting at chunk %d\n", c);
    H = read_group(c, &chunk);
    if (H == NULL) break;

    /*
     * First open new files if required, but not if the current file is still
     * empty.
     */
    if ((SeparateFiles != 0) && (is_splitpoint == 1) && (file_has_data == 1)) {
      if (mfile != NULL) fclose(mfile);
      if (open_mpeg_file(mpeg_name)==0) exit(1);
      SeparateFiles++;
      file_has_data = 0;
    }
    /* Kill the B frames if this is the end of a cut */
    if (i == IS_CUT_END) {
      remove_B_frames(H);
      if (Debug >= 1) {
	printf("After B-removal, "); print_newtylist(H, c);
      }
    }
    /* Now write the thing out */
    if (write_group_as_mpeg(H) == 0) exit(1);

    file_has_data = 1; groups++;
    if ((groups != 0) && ((groups % 20) == 0))
      printf("%3d groups written\n", groups);
  }
  return (0);
}
