/*
 * playitsam: Play, edit & save a TiVo video stream from file or MFS.
 * 
 * readfile.c: Routines that provide a front-end to the various different
 * types of input and output files and the mechanisms used to work on
 * these files.
 *
 * $Revision: 1.7 $
 * 
 * (c) 2002, Warren Toomey wkt@tuhs.org.
 * 
 * Released under the Gnu General Public License version 2
 */

/*
 *
 * Eventually, we will support:
 *
 *	- old input tystreams from Unix files
 *	- old input tystreams from MFS
 *	- old input tystreams from tyserver
 *	- new input tystreams from Unix files
 *	- new input tystreams from tyserver?
 *
 *	- output to new tystream format as Unix file
 *	- output to separated mpa and mpv as Unix files?
 *	- output to SVCD MPEG format as Unix files (drool, drool)
 *
 * Not all of these are implemented as yet.
 */

#include "defines.h"
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include "old_tystream.h"
#include "new_tystream.h"
#include "libplayit.h"

#define INPUT_OLDUNIX		1
#define INPUT_OLDMFS		2
#define INPUT_OLDTYSERVER	3
#define INPUT_NEWUNIX		4
#define INPUT_NEWTYSERVER	5
#define OUTPUT_NEWUNIX		6
#define OUTPUT_SPLIT		7
#define OUTPUT_MPEG		7

static int input_type;		/* Type of input file */
/* static int output_type;		Type of output file */
u_int32_t TotalChunks=0; 	/* Total number of chunks in the old file */
extern int Debug;

/*
 * We keep a cache of recently used 128K chunks. These are tossed in
 * round-robin order.
 * 
 * Chk_cache[] is the cache of actual chunks.
 */
#define CHUNK_CACHE_SIZE        9
static u_int8_t Chk_cache[CHUNK_CACHE_SIZE][CHUNK_SIZE];
static int32_t Chk_cachenum[CHUNK_CACHE_SIZE];
static int new_chunk;           /* Pos'n in array to overwrite data */


				/* Eventually, put all descriptors here */
static int input_fd=0;
FILE *output;
FILE *audio;
FILE *video;


void initialise_chunkcache(void)
{
  int i;
  /* Initialise the cache of chunks */
  for (i=0; i < CHUNK_CACHE_SIZE; i++) {
    Chk_cachenum[i]=-1;  
  }
  new_chunk=0;
}


/* Given a stream name, open that stream and return the
 * number of records in the stream, and the size of the
 * stream in kilobytes. Returns 1 if ok, 0 if failure.
 *
 * The name can be any of the input types above.
 */
int open_new_tystream(char *name, u_int32_t *num_recs, u_int32_t *size_in_k)
{
  char *cptr;

  /* If the name has a colon, must be a tyserver stream */
  if ((cptr=strchr(name,':'))!=NULL) {
    *cptr='\0'; cptr++;
    input_fd= open_mfs_oldtystream(name, cptr, &TotalChunks);
    if (input_fd==-1) return(0);
    initialise_chunkcache();
    *num_recs= TotalChunks;
    *size_in_k= TotalChunks * 128;
    input_type= INPUT_OLDTYSERVER;
    return(1);
  }

  /* For now, try a Unix file */
  input_fd= open_old_local_tystream(name, &TotalChunks);
  if (input_fd==-1) return(0);
  initialise_chunkcache();
  *num_recs= TotalChunks;
  *size_in_k= TotalChunks * 128;
  input_type= INPUT_OLDUNIX;
  return(1);
}


/* Read a chunk in from wherever the file is: MFS, tyserver, file.
 * Returns 1 if ok, 0 on error.
 */
u_int8_t *read_chunk(u_int32_t chunknum)
{
  int i;

  if (chunknum>=TotalChunks) return(NULL);

  /* First, see if the chunk is in the cache. */
  for (i=0; i < CHUNK_CACHE_SIZE; i++) {
    if (Chk_cachenum[i] == chunknum) return(Chk_cache[i]);
  }

  switch (input_type) {
    case INPUT_OLDUNIX:
	i= read_old_local_chunk(input_fd, Chk_cache[new_chunk], chunknum);
	break;
    case INPUT_OLDTYSERVER:
  	i= read_a_chunk(input_fd, chunknum, Chk_cache[new_chunk]);
  }

  if (i==0) return(NULL);
  i= new_chunk;
  new_chunk= (new_chunk+1) % CHUNK_CACHE_SIZE;
  Chk_cachenum[i]= chunknum;
  return(Chk_cache[i]);
}


struct new_ty_node *read_group(u_int32_t start, u_int32_t *next)
{
  switch (input_type) {
    case INPUT_OLDUNIX:
    case INPUT_OLDTYSERVER:
      return(build_group(start, next));
  }
  return(NULL);
}
