/*
 * playitsam: Play, edit & save a TiVo video stream from file or MFS.
 * 
 * readfile.c: Routines to read chunks from various inputs. $Revision: 1.1 $
 * 
 * (c) 2002, Warren Toomey wkt@tuhs.org.
 * 
 * Released under the Gnu General Public License version 2
 */

#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "../libmfs/mfs.h"
#include "tproto.h"


#define CHUNK_SIZE (256*512)		/* Size of a chunk */
#define CHUNKSUNKNOWN	27176654	/* Magic number: indicates that */
					/* we don't know how many chunks */
					/* are in this stream */

#define INPUT_FSID	1		/* TiVo fsid on MFS, or from a */
#define INPUT_FILE	2		/* Unix file  */
#define INPUT_TYSERVER	3		/* Tyserver via TCP */
#define INPUT_TYSERVCHK	4		/* Chunks from tyserver via TCP */

/* playitsam.c now assumes that chunk 0 from every input file contains
 * valid data. For FSIDs, we now ignore chunk 0 always. For other
 * files, we first read chunk 0 and if its is garbage, we set a seek_offset
 * to CHUNK_SIZE to ensure that future seeks wil work.
 */


#ifdef TIVO
/*
 * For multi-fsid recordings, we store the fsids and their sizes here.
 */
struct fsid_group {
  int fsid;
  int start_chunk;
  int end_chunk;
}  *Fsid_list = NULL;
int Fsid_list_count = 0;
#endif

extern int totalchunks;		/* Total chunks */
int inputtype = 0;		/* Type of file input */
off_t seek_offset = 0;		/* Seek needed to skip useless chunk 0 */

#ifdef TIVO
/*
 * Read from a single fsid or from a set of fsids. Always add 1 so as to skip
 * the chunk #0 in every fsid stream.
 */
static int mfs_multifsid_pread(int fsid, char *buf, int chunknum)
{
  int i;

  /* If we only have one fsid, use the one supplied */
  if (Fsid_list_count == 0) {
    return (mfs_fsid_pread(fsid, buf, ((u64) chunknum + 1) << 17, CHUNK_SIZE));
  }

  /* Otherwise, search for an fsid that matches and use it */
  for (i = 0; i < Fsid_list_count; i++) {
    if ((chunknum >= Fsid_list[i].start_chunk) &&
	(chunknum <= Fsid_list[i].end_chunk)) {
      /* fprintf(stderr,"Reading from fsid %d\r\n",Fsid_list[i].fsid); */
      return (mfs_fsid_pread(Fsid_list[i].fsid, buf,
		    ((u64) (1 + chunknum - Fsid_list[i].start_chunk)) << 17,
			     CHUNK_SIZE));
    }
  }
  return (-1);			/* No match */
}
#endif




/* open a socket to a tcp remote host with the specified port */
int open_socket_to_tyserver(char *inhost, int port)
{
  int type = SOCK_STREAM;
  struct sockaddr_in sock_out;
  int res;
  int nodelay = 1;
  int window = 65535;
  char *host, *numptr, *numend;
  u_int32_t raw_address = 0;
  int one_part;

  host = strdup(inhost);
  if (host == NULL) return (-1);

  res = socket(PF_INET, type, 0);
  if (res == -1) { free(host); return (-1); }

#ifdef CAN_DO_NAME_LOOKUPS
  /* Can someone tell me how to get /etc/hosts working on the TiVo? */
  hp = gethostbyname(host);
  if (!hp) {
    fprintf(stderr, "unknown host: %s\n", host);
    return (-1);
  }
  memcpy(&sock_out.sin_addr, hp->h_addr, hp->h_length);

#else

  /* Walk the IP address and convert into a 32-bit integer */
  numptr = host;
  numend = strchr(host, '.');
  if (numend == NULL) {
    fprintf(stderr, "%s not an IP address\n", inhost);
    free(host); return (-1);
  }
  *numend = '\0';
  one_part = atoi(numptr);
  if ((one_part < 0) || (one_part > 255)) {
    fprintf(stderr, "%s not an IP address\n", inhost);
    free(host); return (-1);
  }
  raw_address = one_part << 24;

  numptr = ++numend;
  numend = strchr(numptr, '.');
  if (numptr == NULL) {
    fprintf(stderr, "%s not an IP address\n", inhost);
    free(host); return (-1);
  }
  *numend = '\0';
  one_part = atoi(numptr);
  if ((one_part < 0) || (one_part > 255)) {
    fprintf(stderr, "%s not an IP address\n", inhost);
    free(host); return (-1);
  }
  raw_address += (one_part << 16);

  numptr = ++numend;
  numend = strchr(numptr, '.');
  if (numptr == NULL) {
    fprintf(stderr, "%s not an IP address\n", inhost);
    free(host); return (-1);
  }
  *numend = '\0';
  one_part = atoi(numptr);
  if ((one_part < 0) || (one_part > 255)) {
    fprintf(stderr, "%s not an IP address\n", inhost);
    free(host); return (-1);
  }
  raw_address += (one_part << 8);
  numptr = ++numend;
  one_part = atoi(numptr);
  if ((one_part < 0) || (one_part > 255)) {
    fprintf(stderr, "%s not an IP address\n", inhost);
    free(host); return (-1);
  }
  raw_address += one_part;
  sock_out.sin_addr.s_addr = htonl(raw_address);
#endif

  sock_out.sin_port = htons(port);
  sock_out.sin_family = PF_INET;

  if (connect(res, (struct sockaddr *) & sock_out, sizeof(sock_out))) {
    close(res);
    fprintf(stderr, "failed to connect to %s\n", inhost);
    free(host); return (-1);
  }
  setsockopt(res, SOL_SOCKET, SO_RCVBUF, &window, sizeof(int));
  setsockopt(res, IPPROTO_TCP, TCP_NODELAY, &nodelay, sizeof(int));

  free(host); return (res);
}

/*
 * Read a specific chunk from the input stream. Return 1 if ok, 0 if failure
 */
int read_a_chunk(int fd, int chunknum, char *bptr)
{
  int numread = 0;
  int numleft = CHUNK_SIZE;
  char ch, buf[100];

  switch (inputtype) {
#ifdef TIVO
  case INPUT_FSID:
    numread = mfs_multifsid_pread(fd, bptr, chunknum);
    if (numread != CHUNK_SIZE) {
      fprintf(stderr, "Failure reading chunk %d\r\n", chunknum);
      Exit(1);
    }
    return (1);
#endif

  case INPUT_FILE:
    lseek(fd, seek_offset + (CHUNK_SIZE * chunknum), SEEK_SET);
    goto here;				/* Yuk, a goto */

  case INPUT_TYSERVCHK:
	/* We need to send a CHNK command, get the answer, and then
	 * read the 128K of chunk from the TCP connection.
	 */
      sprintf(buf, "CHNK %d\r\n", chunknum);
      numread = write(fd, buf, strlen(buf));
      if (numread < 1) {
	fprintf(stderr, "Failure sending CHNK command\n");
	Exit(1);
      }
      /* Read the 3-digit result */
      read(fd, buf, 3);
      buf[3] = '\0';
      if (strcmp(buf, "200")) { /* Did not get ok message */
        close(fd);
        fprintf(stderr, "Couldn't read chunk from tyserver\n");
        Exit(1);
      }
      /* Otherwise, read up to \n from tyserver before we get the chunk */
      while (1) {
        read(fd, &ch, 1);
        if (ch == '\n') break;
      }

	/* Fall into the read 128K below */

here:
  case INPUT_TYSERVER:

    /* Loop until we get CHUNK_SIZE bytes of data */
    while (numleft > 0) {
      numread = read(fd, bptr, numleft);
      if (numread < 1) {
	fprintf(stderr, "Failure Reading chunk %d\r\n", chunknum);
	Exit(1);
      }
      numleft -= numread;
      bptr += numread;
    }
    return (1);
  default:
    return (0);
  }
}


/*
 * Given the input argument from the user, work out which one it is and open
 * it. Returns a file descriptor to the input, or a fsid if the input is on
 * MFS. If play_only is true, then we open tyserver streams with PLAY, and
 * we don't use the CHNK command.
 */
int open_input(char *inputsource, int play_only)
{
#ifdef TIVO
  int fsid;
  int i;
  char *onefsid;
#endif
  int fd;
  struct stat statbuf;
  char ch, *tyfile;
  char result[4];
  unsigned char buf[4];

  /* If the input source has a colon, get it from the tyserver */
  if ((play_only==1) && ((tyfile = strchr(inputsource, ':')) != NULL)) {
    *tyfile = '\0';
    tyfile++;
    fd = open_socket_to_tyserver(inputsource, 1150);
    if (fd == -1)
      Exit(1);

    /* Now send a PLAY command, and get a 200 reply */
    write(fd, "PLAY ", 5);
    write(fd, tyfile, strlen(tyfile));
    write(fd, "\r\n", 2);
    read(fd, result, 3);
    result[3] = '\0';
    if (strcmp(result, "200")) {/* Did not get ok message */
      close(fd);
      fprintf(stderr, "Couldn't play from tyserver\n");
      Exit(1);
    }
    /* Otherwise, read up to \n from tyserver before we get the stream */
    while (1) {
      read(fd, &ch, 1);
      if (ch == '\n')
	break;
    }
    totalchunks = CHUNKSUNKNOWN;
    inputtype = INPUT_TYSERVER;
    return (fd);
  }

  /* If the input source has a colon, get it from the tyserver */
  /* This version fo the code is for reading chunks 1 at a time. */
  if ((play_only==0) && ((tyfile = strchr(inputsource, ':')) != NULL)) {
    *tyfile = '\0';
    tyfile++;
    fd = open_socket_to_tyserver(inputsource, 1150);
    if (fd == -1)
      Exit(1);

    /* Now send an OPEN command, and get a 200 reply */
    write(fd, "OPEN ", 5);
    write(fd, tyfile, strlen(tyfile));
    write(fd, "\r\n", 2);
    read(fd, result, 3);
    result[3] = '\0';
    if (strcmp(result, "200")) { /* Did not get ok message */
      close(fd);
      fprintf(stderr, "Couldn't play from tyserver\n");
      Exit(1);
    }

    /* Read the size of the tystream */
    read(fd, &ch, 1);	/* Read the space */
    totalchunks=0;
    while (1) {
      read(fd, &ch, 1);		/* Read a digit or space */
      if (ch == ' ') break;
      totalchunks= 10 * totalchunks + (ch - '0');
    }
    fprintf(stderr, "Just got %d chunks back from the tyserver!!!\n",totalchunks);

    /* Now read up to \n from tyserver before we get the stream */
    while (1) {
      read(fd, &ch, 1);
      if (ch == '\n')
	break;
    }
    inputtype = INPUT_TYSERVCHK;
    return (fd);
  }

#ifdef TIVO
#define MAX_FSID_LIST 50

  /**** BIG CHANGE FROM PREVOUS VERSIONS ****/
  /* We now always ignore chunk 0 from every fsid stream */

  /* inputsource could be a single fsid or a list of fsids */
  /* First, check if it has commas in it */
  /* If so, split the list into fsids, store them */
  /* and work out their size */
  if (strchr(inputsource, ',')) {

    /* Malloc room for MAX_FSID_LIST entries */
    Fsid_list = (struct fsid_group *) malloc(
				 MAX_FSID_LIST * sizeof(struct fsid_group));
    if (Fsid_list == NULL) {
      fprintf(stderr, "Can't malloc fsid list\n");
      exit(1);
    }
    mfs_init();
    totalchunks = 0;
    fsid = 0;
    /* Extract each fsid */
    for (i=0;(i<MAX_FSID_LIST)&&((onefsid=strsep(&inputsource,","))!=NULL);i++){
      fsid = atoi(onefsid);
      if (fsid <= 0)
	continue;
      Fsid_list[i].fsid = fsid;
      Fsid_list[i].start_chunk = totalchunks;
      totalchunks += mfs_fsid_numchunks(fsid) - 2;   /* Don't count chunk #0 */
						/* Nor last chunk */
      Fsid_list[i].end_chunk = totalchunks - 1;
      Fsid_list_count++;
#if 0
      fprintf(stderr, "fsid %d has %d chunks [%d to %d]\r\n", fsid, totalchunks,
	      Fsid_list[i].start_chunk, Fsid_list[i].end_chunk);
#endif
    }
    inputtype = INPUT_FSID;
    return (fsid);
  }

  /* Nope, no commas, but if the input source is numeric, treat it as a fsid */
  fsid = atoi(inputsource);
  if (fsid != 0) {
    mfs_init();
    totalchunks = mfs_fsid_numchunks(fsid) - 1;	/* Skip chunk 0 */
    fprintf(stderr, "fsid %d has %d chunks\r\n", fsid, totalchunks);
    inputtype = INPUT_FSID;
    return (fsid);
  }
#endif

  /* None of the above, treat it as a Unix filename */
  fd = open(inputsource, O_RDONLY);
  if (fd == -1) {
    fprintf(stderr, "Cannot open input file %s\n", inputsource);
    Exit(1);
  }
  if (stat(inputsource, &statbuf) == -1) {
    fprintf(stderr, "Cannot determine size of %s\n", inputsource);
    Exit(1);
  }

  /* Read the first chunk. If it doesn't contain F5 46 XX BD, then it's
   * a valid chunk, else skip it.
   */
  read(fd, buf, 4);
  if ((buf[0]!=0xf5) && (buf[1]!=0x46) && (buf[3]!=0xbd)) seek_offset=0;
  else seek_offset= CHUNK_SIZE;
  
  totalchunks = (statbuf.st_size - seek_offset) / CHUNK_SIZE;
  inputtype = INPUT_FILE;
  return (fd);
}
