/*
 * playitsam: Play, edit & save a TiVo video stream from file or MFS.
 * 
 * ttyserver.c: Tystream server for TiVo only. $Revision: 1.6 $
 * 
 * (c) 2002, Warren Toomey wkt@tuhs.org.
 * 
 * Released under the Gnu General Public License version 2
 */

#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include "tproto.h"

/* Why isn't this defined in TiVo Linux? */
int snprintf(char *str, size_t size, const char *format,...);


#define CHUNK_SIZE (256*512)
char buf[CHUNK_SIZE];
int totalchunks;		/* Total chunks */
int infile = 0;			/* FD of input file */
int chunk_in_buf = -1;		/* # of the chunk in the buffer */

void Exit(int num)
{
  exit(num);
}

void exit_server(int sockfd)
{
  char *okmsg = "200 See you later\r\n";
  write(sockfd, okmsg, strlen(okmsg));
  close(sockfd);
  exit(0);
}

void play_file(int sockfd, char *filename)
{
  int chk;
  int numleft, numsent;
  unsigned char *bptr;

  char *errmsg1 = "400 Cannot open requested stream\r\n";
  char *okmsg = "200 Ok, here is the stream\r\n";

  infile = open_input(filename, 0);
  if (infile == -1) {
    write(sockfd, errmsg1, strlen(errmsg1));
    close(sockfd); exit(1);
  }
  write(sockfd, okmsg, strlen(okmsg));

  for (chk = 0; chk < totalchunks; chk++) {

    read_a_chunk(infile, chk, buf);

    /* Write the chunk to the output fd */
    numleft = CHUNK_SIZE;
    bptr = buf;
    while (numleft > 0) {
      numsent = write(sockfd, bptr, numleft);
      if (numsent < 1) exit(1);
      numleft -= numsent;
      bptr += numsent;
    }
  }
  close(sockfd);
  exit(0);
}

void open_file(int sockfd, char *filename)
{
  char *errmsg1 = "400 Cannot open requested stream\r\n";
  char okbuf[100];

  infile = open_input(filename, 0);
  if (infile == -1) write(sockfd, errmsg1, strlen(errmsg1));
  else {
    snprintf(okbuf, 98, "200 %d chunks in this stream\r\n", totalchunks);
    write(sockfd, okbuf, strlen(okbuf));
  }

  /* We assume that the first chunk to be used is chunk 0 */
  read_a_chunk(infile, 0, buf);
  chunk_in_buf = 0;
  return;
}

void send_chunk(int sockfd, int chk)
{
  char *okmsg = "200 Ok, here is the chunk\r\n";

  int numleft, numsent;
  unsigned char *bptr;

  /* Read the chunk if we haven't already cached it! */
  if (chunk_in_buf != chk) {
    read_a_chunk(infile, chk, buf);
    chunk_in_buf = chk;
  }
  write(sockfd, okmsg, strlen(okmsg));

  /* Write the chunk to the output fd */
  numleft = CHUNK_SIZE;
  bptr = buf;
  while (numleft > 0) {
    numsent = write(sockfd, bptr, numleft);
    if (numsent < 1) exit(1);
    numleft -= numsent;
    bptr += numsent;
  }

  /* And cache the next chunk */
  read_a_chunk(infile, ++chunk_in_buf, buf);
}


void handle_request(int sockfd)
{
#define BUFSIZE 1024
  char *errmsg1 = "400 Unrecognised command\r\n";
  char cmd[BUFSIZE];
  int rval;
  int i;

  /*
   * Commands are single lines ending in \r\n. Only 4 commands:
   * PLAY ffff\r\n   to play the file ffff. No slashes in ffff
   * OPEN ffff\r\n   to prepare the use of a file for chunk playing
   * CHNK nnnn\r\n   to play the 128K chunk from file, nnnn is decimal
   * EXIT	     to exit
   * 
   * Server always returns 1 response line, either
   * 200 Ok\r\n     or
   * 400 Error\r\n
   * Client should read 3 digits & ignore the rest.
   * 
   * PLAY returns a tystream until it ends or the connection is broken.
   * Once the stream is sent, the server dies and the connection closes.
   * 
   * OPEN returns nothing but the 200 or 400 response line. The server stays
   * alive waiting for the next command.
   * 
   * CHNK returns the the 200 or 400 response line followed by 128K of chunk.
   * The server stays alive waiting for the next command.
   * 
   * EXIT returns a 200 message, and the server dies.
   */

  /* Read a command from the socket */
  /* Ok, so I should read a whole buffer in one hit, who cares! */
  for (i = 0; i < BUFSIZE - 2; i++) {
    rval = read(sockfd, &cmd[i], 1);
    if ((rval == -1) || (cmd[i] == '\n')) break;
  }
  cmd[++i] = '\0';		/* Null-terminate the command */

  if (!strncmp(cmd, "EXIT", 4)) { exit_server(sockfd); return; }

  /* Walk the command and remove \r and \n */
  for (i = 0; i < BUFSIZE; i++) {
    if ((cmd[i] == '\r') || (cmd[i] == '\n')) { cmd[i] = '\0'; break; }
  }

  if (!strncmp(cmd, "PLAY", 4)) { play_file(sockfd, &cmd[5]); return; }
  if (!strncmp(cmd, "OPEN", 4)) { open_file(sockfd, &cmd[5]); return; }
  if (!strncmp(cmd, "CHNK", 4)) { send_chunk(sockfd, atoi(&cmd[5])); return; }
  write(sockfd, errmsg1, strlen(errmsg1));
  return;
}

void usage(void)
{
  printf("Usage: ttyserver\n"); exit(1);
}

int main(int argc, char *argv[])
{
  int sock, length;
  struct sockaddr_in server;
  int pid, msgsock;
  int nodelay = 1;			/* Turn on TCP_NODELAY */

  if (argc != 1) usage();

  /* Create socket */
  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    perror("opening stream socket"); exit(1);
  }

  /* Name socket using wildcards */
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;

  /*
   * Why 1150? T looks like 1, i looks like 1 V is 5 in roman numerals, o
   * looks like 0
   */
  server.sin_port = ntohs(1150);
  if (bind(sock, (struct sockaddr *) & server, sizeof(server))) {
    perror("binding stream socket"); exit(1);
  }

  /* Find out assigned port number and print it out */
  length = sizeof(server);
  if (getsockname(sock, (struct sockaddr *) & server, &length)) {
    perror("getting socket name"); exit(1);
  }

  /* Print out that we have the socket, and become a daemon */
  printf("Socket has port #%d\n", ntohs(server.sin_port));
  daemon(0,0);

  /* Start accepting connections */
  listen(sock, 5);

  while (1) {
    msgsock = accept(sock, 0, 0);
    if (msgsock == -1)
      perror("accept");

    if (nodelay==1)
      setsockopt(msgsock, IPPROTO_TCP, TCP_NODELAY, &nodelay, sizeof(int));

#define FORK
#ifdef FORK
    pid = fork();
    if (pid == 0)
      while (1) handle_request(msgsock);
    else {
      close(msgsock);
      /* Clean up as many children as we can */
      waitpid(-1, &length, WNOHANG);
      waitpid(-1, &length, WNOHANG);
      waitpid(-1, &length, WNOHANG);
      waitpid(-1, &length, WNOHANG);
    }
#else
    while (1) handle_request(msgsock);
#endif
  }
}
